/*
 * Copyright 1997-2013 Optimatika (www.optimatika.se)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.ojalgo.finance.portfolio;

import java.math.BigDecimal;

import org.ojalgo.ProgrammingError;
import org.ojalgo.array.ArrayUtils;
import org.ojalgo.constant.BigMath;
import org.ojalgo.constant.PrimitiveMath;
import org.ojalgo.finance.FinanceUtils;
import org.ojalgo.matrix.BasicMatrix;
import org.ojalgo.matrix.PrimitiveMatrix;
import org.ojalgo.matrix.decomposition.Eigenvalue;
import org.ojalgo.matrix.decomposition.EigenvalueDecomposition;
import org.ojalgo.matrix.store.MatrixStore;
import org.ojalgo.matrix.store.PhysicalStore;
import org.ojalgo.scalar.BigScalar;
import org.ojalgo.scalar.Scalar;
import org.ojalgo.type.TypeUtils;

/**
 * MarketEquilibrium translates between the market portfolio weights
 * and the equilibrium excess returns. The only things needed to do
 * those translations are the covariance matrix and the risk aversion
 * factor - that's what you need to supply when you instantiate this
 * class.
 *
 * @see #calculateAssetReturns(BasicMatrix)
 * @see #calculateAssetWeights(BasicMatrix)
 * 
 * @author apete
 */
public class MarketEquilibrium {

    private static final BigDecimal DEFAULT_RISK_AVERSION = BigMath.ONE; // Don't change the default!
    private static final String STRING_ZERO = "0";
    private static final String SYMBOL = "Asset_";

    /**
     * Calculates the portfolio return using the input instrument
     * weights and returns.
     */
    public static Scalar<?> calculatePortfolioReturn(final BasicMatrix aWeightsVctr, final BasicMatrix aReturnsVctr) {
        return aWeightsVctr.multiplyVectors(aReturnsVctr);
    }

    private static String[] makeSymbols(final int aCount) {

        final String[] retVal = new String[aCount];

        final int tmpMaxLength = Integer.toString(aCount - 1).length();

        String tmpNumberString;
        for (int i = 0; i < aCount; i++) {
            tmpNumberString = Integer.toString(i);
            while (tmpNumberString.length() < tmpMaxLength) {
                tmpNumberString = STRING_ZERO + tmpNumberString;
            }
            retVal[i] = SYMBOL + tmpNumberString;
        }

        return retVal;
    }

    private final BasicMatrix myCovariances;
    private BigDecimal myRiskAversion;
    private final String[] myAssetKeys;

    public MarketEquilibrium(final BasicMatrix aCovarianceMatrix) {
        this(aCovarianceMatrix, DEFAULT_RISK_AVERSION);
    }

    public MarketEquilibrium(final BasicMatrix aCovarianceMatrix, final Number aRiskAversionFactor) {
        this(MarketEquilibrium.makeSymbols(aCovarianceMatrix.getRowDim()), aCovarianceMatrix, aRiskAversionFactor);
    }

    public MarketEquilibrium(final String[] assetNamesOrKeys, final BasicMatrix aCovarianceMatrix) {

        super();

        myAssetKeys = ArrayUtils.copyOf(assetNamesOrKeys);
        myCovariances = aCovarianceMatrix;
        myRiskAversion = DEFAULT_RISK_AVERSION;
    }

    public MarketEquilibrium(final String[] assetNamesOrKeys, final BasicMatrix aCovarianceMatrix, final Number aRiskAversionFactor) {

        super();

        myAssetKeys = ArrayUtils.copyOf(assetNamesOrKeys);
        myCovariances = aCovarianceMatrix;
        myRiskAversion = TypeUtils.toBigDecimal(aRiskAversionFactor);
    }

    @SuppressWarnings("unused")
    private MarketEquilibrium() {

        this(null, null, null);

        ProgrammingError.throwForIllegalInvocation();
    }

    MarketEquilibrium(final MarketEquilibrium aMarket) {
        this(aMarket.getAssetKeys(), aMarket.getCovariances(), aMarket.getRiskAversion().getNumber());
    }

    /**
     * If the input vector of asset weights are the weights of the market
     * portfolio, then the ouput is the equilibrium excess returns.
     */
    public BasicMatrix calculateAssetReturns(final BasicMatrix aWeightsVctr) {
        final BasicMatrix tmpWeightsVctr = myRiskAversion.compareTo(DEFAULT_RISK_AVERSION) == 0 ? aWeightsVctr : aWeightsVctr.multiply(myRiskAversion);
        return myCovariances.multiplyRight(tmpWeightsVctr);
    }

    /**
     * If the input vector of returns are the equilibrium excess returns
     * then the output is the market portfolio weights. This is
     * unconstrained optimisation - there are no constraints on the
     * resulting instrument weights.
     */
    public BasicMatrix calculateAssetWeights(final BasicMatrix aReturnsVctr) {
        final BasicMatrix tmpWeightsVctr = myCovariances.solve(aReturnsVctr);
        if (myRiskAversion.compareTo(DEFAULT_RISK_AVERSION) == 0) {
            return tmpWeightsVctr;
        } else {
            return tmpWeightsVctr.divide(myRiskAversion);
        }
    }

    /**
     * Calculates the portfolio variance using the input instrument
     * weights.
     */
    public Scalar<?> calculatePortfolioVariance(final BasicMatrix aWeightsVctr) {

        BasicMatrix tmpLeft;
        BasicMatrix tmpRight;

        if (aWeightsVctr.getColDim() == 1) {
            tmpLeft = aWeightsVctr.transpose();
            tmpRight = aWeightsVctr;
        } else {
            tmpLeft = aWeightsVctr;
            tmpRight = aWeightsVctr.transpose();
        }

        return myCovariances.multiplyRight(tmpRight).multiplyLeft(tmpLeft).toScalar(0, 0);
    }

    /**
     * Will set the risk aversion factor to the best fit for an observed
     * pair of market portfolio asset weights and equilibrium/historical
     * excess returns.
     */
    public void calibrate(final BasicMatrix aWeightsVctr, final BasicMatrix aReturnsVctr) {

        final Scalar<?> tmpImpliedRiskAversion = this.calculateImpliedRiskAversion(aWeightsVctr, aReturnsVctr);

        this.setRiskAversion(tmpImpliedRiskAversion.getNumber());
    }

    /**
     * Equivalent to copying, but additionally the covariance matrix will be cleaned
     * of negative and very small eigenvalues to make it positive definite.
     */
    public MarketEquilibrium clean() {

        final Eigenvalue<Double> tmpEvD = EigenvalueDecomposition.makePrimitive(true);
        tmpEvD.compute(myCovariances, false);

        final MatrixStore<Double> tmpV = tmpEvD.getV();
        final PhysicalStore<Double> tmpD = tmpEvD.getD().copy();

        final double tmpLargest = tmpD.doubleValue(0, 0);
        final double tmpLimit = Math.max(PrimitiveMath.MACHINE_DOUBLE_ERROR * tmpLargest, 1E-12);

        for (int ij = 0; ij < tmpD.getMinDim(); ij++) {
            if (tmpD.doubleValue(ij, ij) < tmpLimit) {
                tmpD.set(ij, ij, tmpLimit);
            }
        }

        final MatrixStore<Double> tmpLeft = tmpV;
        final MatrixStore<Double> tmpMiddle = tmpD;
        final MatrixStore<Double> tmpRight = tmpLeft.builder().transpose().build();

        final MatrixStore<Double> tmpCovar = tmpMiddle.multiplyLeft(tmpLeft).multiplyRight(tmpRight);
        final BasicMatrix tmpCovariances = PrimitiveMatrix.FACTORY.copy(tmpCovar);

        return new MarketEquilibrium(myAssetKeys, tmpCovariances, myRiskAversion);
    }

    public MarketEquilibrium copy() {
        return new MarketEquilibrium(this);
    }

    public String getAssetKey(final int index) {
        return myAssetKeys[index];
    }

    public String[] getAssetKeys() {
        return ArrayUtils.copyOf(myAssetKeys);
    }

    public BasicMatrix getCovariances() {
        return myCovariances;
    }

    public Scalar<?> getRiskAversion() {
        return new BigScalar(myRiskAversion);
    }

    /**
     * @deprecated v35 Use {@link #getAssetKeys()} or {@link #getAssetKey(int)} instead.
     */
    @Deprecated
    public String[] getSymbols() {
        return this.getAssetKeys();
    }

    public void setRiskAversion(final Number aFactor) {

        final BigDecimal tmpFactor = TypeUtils.toBigDecimal(aFactor);

        if (tmpFactor.signum() == 0) {
            myRiskAversion = DEFAULT_RISK_AVERSION;
        } else if (tmpFactor.signum() < 0) {
            myRiskAversion = tmpFactor.negate();
        } else {
            myRiskAversion = tmpFactor;
        }
    }

    public int size() {
        return Math.min(myCovariances.getRowDim(), myCovariances.getColDim());
    }

    public BasicMatrix toCorrelations() {
        return FinanceUtils.toCorrelations(myCovariances, false);
    }

    /**
     * Will calculate the risk aversion factor that is the best fit for
     * an observed pair of market portfolio weights and equilibrium/historical
     * excess returns.
     */
    Scalar<?> calculateImpliedRiskAversion(final BasicMatrix<?> aWeightsVctr, final BasicMatrix<?> aReturnsVctr) {

        Scalar<?> retVal = myCovariances.multiplyRight(aWeightsVctr).solve(aReturnsVctr).toScalar(0, 0);

        if (retVal.isZero()) {
            retVal = BigScalar.ONE;
        } else if (!retVal.isPositive()) {
            retVal = retVal.negate();
        }

        return retVal;
    }

}
